package Agent ;

import java.lang.reflect.* ;
import java.io.* ;
import java.net.* ;
import java.util.* ;

public abstract class AgentEx extends Agent {
  protected Hashtable usedClasses = null ; // ( name, class bytes )

  public AgentEx( String s ) {
    super( s ) ;
  }

  public AgentEx() {
    super() ;
  }

  synchronized public byte[] getClassBytes( String className ) {
    byte[] ClassBytes = null ;

    if ( usedClasses != null )
      ClassBytes = (byte[])usedClasses.get( className ) ;

    if ( ClassBytes != null )
      return ClassBytes ;
    else
      ClassBytes = super.getClassBytes( className ) ;

    return ClassBytes ;
  }

  synchronized protected AgentPacket createAgentPacket() {
    return new AgentPacketEx( this, getClassBytes(), getUsedClasses() ) ;
  }

  synchronized protected Hashtable getUsedClasses() {
    if ( usedClasses == null )
      createUsedClassesTable() ;

    return usedClasses ;
  }

  synchronized protected void createUsedClassesTable() {
    if ( usedClasses != null )
      return ;

    usedClasses = new Hashtable() ;
    Vector UsedClassesNames = new Vector() ;

    // otteniamo "tutte" le classi utilizzate da questa classe
    getUsedClassesNames( getClass(), UsedClassesNames ) ;
    byte[] classBytes ;
    String className ;
    Enumeration en = UsedClassesNames.elements() ;
    while ( en.hasMoreElements() ) {
      className = (String)en.nextElement() ;
      //PrintMessage( "classe " + className + " ...") ;
      classBytes = getClassBytes( className ) ;
      if ( classBytes != null ) {
      	usedClasses.put( className, classBytes ) ;
	      PrintMessage( "Registrata " + className ) ;
      }
    }
  }

  public Enumeration getUsedClassesNames() {
    if ( usedClasses != null )
      return usedClasses.keys() ;

    return null ;
  }

  // tramite le reflection API si ottengono le classi dei vari membri
  // e dei parametri e dei valori di ritorno dei vari metodi della
  // classe specificata
  protected void getUsedClassesNames( Class c, Vector result ) {
    Field[] fields = c.getDeclaredFields() ;
    Constructor[] constructors = c.getDeclaredConstructors() ;
    Method[] methods = c.getDeclaredMethods() ;
    Class[] declClasses = c.getDeclaredClasses() ;
    Class[] classes ;
    int i, j ;

    for( i = 0 ; i < fields.length ; i++ ) {
      getUsedClassesNamesOf( fields[i].getType(), result ) ;
    }

    for ( i = 0; i < constructors.length; i++ ) {
      classes = constructors[i].getParameterTypes() ;
      if ( classes.length > 0 ) {
      	for ( j = 0; j < classes.length; j++ )
	        getUsedClassesNamesOf( classes[j], result ) ;
      }
    }

    for ( i = 0; i < methods.length; i++ ) {
      getUsedClassesNamesOf( methods[i].getReturnType(), result ) ;
      classes = methods[i].getParameterTypes() ;
      if ( classes.length > 0 ) {
        for ( j = 0; j < classes.length; j++ )
          getUsedClassesNamesOf( classes[j], result ) ;
      }
    }

    // purtroppo getDeclaredClasses non  ancora implementata nel jdk
    // quindi questi  in attesa che venga implementata, per adesso
    // e' inutile (il vettore e' vuoto)   :-(
    for ( i = 0; i < declClasses.length; i++ ) {
      getUsedClassesNamesOf( declClasses[i], result ) ;
    }

    getUsedClassesNamesOf( c.getSuperclass(), result ) ;
  }

  protected void getUsedClassesNamesOf( Class classVar, Vector result ) {
    String className = filter( classVar.getName() ) ;
    if ( isUsefulClass( className ) && ! result.contains( className ) ) {
      // e' la prima volta
      result.addElement( className ) ;
      // chiamata ricorsiva
      getUsedClassesNames( classVar, result ) ;
    }
  }

  // ripulisce da caratteri indesiderati (se e' un array ad esempio)
  protected String filter( String className ) {
    if ( className.startsWith( "[L" ) )
      return className.substring( 2 ) ;
    return className ;
  }

  // E' davvero necessario spedire questa classe ?
  final protected boolean isUsefulClass( String className )
  {
    boolean result ;
    result = (
      className.equals( "void" ) ||
      className.equals( "int" ) ||
      className.equals( "char" ) ||
      className.equals( "double" ) ||
      className.equals( "float" ) ||
      className.equals( "long" ) ||
      className.equals( "short" ) ||
      className.equals( "boolean" ) ||
      className.equals( "byte" ) ||
      (
        className.indexOf( '.' ) != -1 &&
        (className.startsWith("java") || className.startsWith("Agent"))
      )
    ) ;
    return (! result) ;
  }
}

